// 
// BoardTests.cs
//  
// Author:
//       Matteo Nardi <matteo.nardi@gmx.com>
// 
// Copyright (c) 2011 Matteo Nardi
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Text;
using NUnit.Framework;

namespace Draughts.Tests
{
	[TestFixture]
	public class BoardTests
	{
		Board board;
		Draughts piece;
		BoardPosition position;
		
		[SetUp]
		public void SetUp ()
		{
			int rows = 8;
			int columns = 8;
			board = new Board (rows, columns);
			
			piece = Draughts.BlackMan;
			position = new BoardPosition (5, 4);
		}
		
		[Test]
		public void AllPositionsProperty_IteratesAllValidPositions ()
		{
			int rows = 3;
			int columns = 2;
			board = new Board (rows, columns);
			StringBuilder str = new StringBuilder ();
			
			foreach (BoardPosition position in board.AllPositions)
				str.Append (position.ToString ());
			
			string expected = "[0-0][0-1][1-0][1-1][2-0][2-1]";
			Assert.AreEqual (expected, str.ToString());
		}
		
		[Test]
		public void ContainsPieceIn_EmptyPosition_ReturnsFalse ()
		{
			Assert.IsFalse (board.ContainsPieceIn (position));
		}
		
		[Test]
		public void ContainsPieceIn_NonEmptyPosition_ReturnsTrue ()
		{
			board.PutPieceIn (position, piece);
			
			Assert.IsTrue (board.ContainsPieceIn (position));
		}
		
		[Test]
		public void PieceIn_NonEmptyPosition_ReturnsRightPiece ()
		{
			board.PutPieceIn (position, piece);
			
			Assert.AreEqual (piece, board.PieceIn (position));
		}
		
		[Test]
		[ExpectedException (typeof (InvalidOperationException),
		 ExpectedMessage = "Can't add piece to non empty position [5-4]")]
		public void PutPieceIn_NonEmptyPosition_ThrowsException ()
		{
			position = new BoardPosition (5, 4);
			board.PutPieceIn (position, piece);
			board.PutPieceIn (position, piece);
		}
		
		[Test]
		public void RemovePieceIn_NonEmptyPosition_RemovesPiece ()
		{
			board.PutPieceIn (position, piece);
			board.RemovePieceIn (position);
			
			Assert.IsFalse (board.ContainsPieceIn (position));
		}
		
		[Test]
		[ExpectedException (typeof (InvalidOperationException),
		 ExpectedMessage = "Can't remove piece from empty position [5-4]")]
		public void RemovePieceIn_EmptyPosition_ThrowsException ()
		{
			position = new BoardPosition (5, 4);
			board.RemovePieceIn (position);
		}
		
		[Test]
		public void ContainsPosition_WithValidPosition_ReturnsTrue ()
		{
			foreach (BoardPosition position in board.AllPositions)
				Assert.IsTrue (board.ContainsPosition (position));
		}
		
		[Test]
		[ExpectedException (typeof (InvalidOperationException),
		 ExpectedMessage = "In Draughts Boards you are allowed to put pieces only on dark squares")]
		public void PutPieceIn_LightSquare_ThrowsException ()
		{
			position = new BoardPosition (0, 0);
			board.PutPieceIn (position, Draughts.BlackKing);
		}
	}
}

