// 
// GameTests.cs
//  
// Author:
//       Matteo Nardi <matteo.nardi@gmx.com>
// 
// Copyright (c) 2011 Matteo Nardi
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using NUnit.Framework;

namespace Draughts.Tests
{
	[TestFixture]
	public class GameTests
	{
		MoveDescription genericMove;
		Board board;
		PlayerMock black;
		PlayerMock white;
		GameMock game;
		
		[SetUp]
		public void SetUp ()
		{
			board = Board.BuildDefault ();
			black = new PlayerMock ();
			white = new PlayerMock ();
			BoardPosition genericPosition = new BoardPosition ();
			genericMove = new MoveDescription (genericPosition, genericPosition);
			
			game = new GameMock (board, white, black);
			game.BoardAnalysisMock.GetAllMovesAnswer = new List <MoveDescription> { genericMove, genericMove };
		}
		
		[Test]
		public void RetriveNewMove_OnStoppedGame_AsksPlayerToProduceMove ()
		{
			game.RetriveNewMove ();
			Assert.IsTrue (white.IsProducingMove);
		}
		
		[Test]
		[ExpectedException (typeof (InvalidOperationException),
		 ExpectedMessage = "Wrong player tried to move!")]
		public void OnNewMoveDecided_ByWrongPlayer_ExceptionIsThrown ()
		{
			game.RetriveNewMove ();
			black.Move (genericMove);
		}
		
		[Test]
		[ExpectedException (typeof (InvalidOperationException),
		 ExpectedMessage = "Player tried to perform illegal move!")]
		public void OnNewMoveDecided_InvalidMove_ExceptionIsThrown ()
		{
			game.RetriveNewMove ();
			
			// Instruct stub to treat move as invalid
			game.BoardAnalysisMock.IsValidMoveAnswer = false;
			
			white.Move (genericMove);
		}
		
		[Test]
		public void OnNewMoveDecided_CorrectMove_ActionIsPerformed ()
		{
			MoveActionStub action = game.BoardAnalysisMock.GetMoveActionAnswer;
			
			game.RetriveNewMove ();
			white.Move (genericMove);
			
			Assert.IsTrue (action.Executed);
		}
		
		[Test]
		public void Status_AfterConstructor_IsStopped ()
		{
			Assert.AreEqual (GameStatus.Stopped, game.Status);
		}
		
		[Test]
		public void Status_AfterNewMove_IsStopped ()
		{
			game.RetriveNewMove ();
			white.Move (genericMove);
			
			Assert.AreEqual (GameStatus.Stopped, game.Status);
		}
		
		[Test]
		public void Status_AfterRetriveNewMove_IsWaitingForMove ()
		{
			game.RetriveNewMove ();
			
			Assert.AreEqual (GameStatus.WaitingForMove, game.Status);
		}
	}
	
	class PlayerMock : IPlayer
	{
		public bool IsProducingMove = false;
		
		public void StartGetMove () 
		{
			IsProducingMove = true;
		}
		
		public void Move (MoveDescription move)
		{
			MoveDecided (this, move);
			IsProducingMove = false;
		}
		
		public event Action <IPlayer, MoveDescription> MoveDecided;
	}
	
	class GameMock : Game
	{
		public BoardAnalysisMock BoardAnalysisMock = new BoardAnalysisMock ();
		
		public GameMock (Board board, IPlayer white, IPlayer black)
			: base (board, white, black)
		{
		}
		
		protected override IBoardAnalysis GetBoardAnalysis ()
		{
			return BoardAnalysisMock;
		}
	}
	
	class BoardAnalysisMock : IBoardAnalysis
	{
		public bool IsValidMoveAnswer = true;
		public MoveActionStub GetMoveActionAnswer = new MoveActionStub ();
		public List <MoveDescription> GetAllMovesAnswer = new List <MoveDescription> ();
		
		public bool IsValidMove (MoveDescription move)
		{
			return IsValidMoveAnswer;
		}
		
		public List <MoveDescription> GetAllMoves ()
		{
			return GetAllMovesAnswer;
		}
		
		public IMoveAction GetMoveAction (MoveDescription move)
		{
			return GetMoveActionAnswer;
		}
	}
	
	class MoveActionStub : IMoveAction
	{
		public bool Executed = false;
		public bool Undone = false;
				
		public void Execute (GenericBoard <Draughts> board)
		{
			Executed = true;
		}
		
		public void Undo (GenericBoard <Draughts> board)
		{
			Undone = true;
		}
	}
}

