// 
// MoveActionTests.cs
//  
// Author:
//       Matteo Nardi <matteo.nardi@gmx.com>
// 
// Copyright (c) 2011 Matteo Nardi
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using NUnit.Framework;

namespace Draughts.Tests
{
	[TestFixture]
	public class MoveActionTests
	{
		
		Board board;
		BoardPosition position0, position1, position2;
		Draughts piece0, piece1, piece2, piece3;
		
		List <BoardPosition> locationsToBeEmptied;
		List <Draughts> removedPieces;
		List <BoardPosition> locationsToBeFilled;
		List <Draughts> addedPieces;
		
		MoveAction move;
		
		[SetUp]
		public void SetUp ()
		{
			board = new Board (8, 8);
			
			// Generic positions
			position0 = new BoardPosition (3, 2);
			position1 = new BoardPosition (7, 2);
			position2 = new BoardPosition (8, 1);
			
			// Generic Pieces
			piece0 = Draughts.BlackKing;
			piece1 = Draughts.BlackMan;
			piece2 = Draughts.WhiteKing;
			piece3 = Draughts.WhiteMan;
			
			locationsToBeEmptied = new List <BoardPosition> ();
			removedPieces = new List <Draughts> ();
			locationsToBeFilled = new List <BoardPosition> ();
			addedPieces = new List <Draughts> ();
		}
		
		MoveAction BuildMoveAction () 
		{
			return new MoveAction (locationsToBeEmptied,
			                       removedPieces,
			                       locationsToBeFilled,
			                       addedPieces);
		}
		
		void SetUpMoveActionToRemovePieceIn (BoardPosition position)
		{
			locationsToBeEmptied.Add (position);
			removedPieces.Add (board.PieceIn (position));
		}
		
		void SetUpMoveActionToPutPieceIn (BoardPosition position, Draughts piece)
		{
			locationsToBeFilled.Add (position);
			addedPieces.Add (piece);
		}

		[Test]
		[ExpectedException (typeof (ArgumentException),
		 ExpectedMessage = "locationsToBeEmptied has different number of elements then removedPieces")]
		public void Constructor_OnNotMatchingNumberOfRemovedPiecesAndLocationsToBeEmptied_ThrowsException ()
		{
			locationsToBeEmptied.Add (position0);
			
			move = BuildMoveAction ();
		}
		
		[Test]
		[ExpectedException (typeof (ArgumentException),
		 ExpectedMessage = "locationsToBeFilled has different number of elements then addedPieces")]
		public void Constructor_OnNotMatchingNumberOfAddedPiecesAndLocationsToBeFilled_ThrowsException ()
		{
			addedPieces.Add (piece0);
			
			move = BuildMoveAction ();
		}
		
		[Test]
		public void Execute_RemovingPieces_RemovesPieces ()
		{
			board.PutPieceIn (position0, piece0);
			board.PutPieceIn (position1, piece1);
			
			SetUpMoveActionToRemovePieceIn (position0);
			SetUpMoveActionToRemovePieceIn (position1);
			move = BuildMoveAction ();
			move.Execute (board);
			
			Assert.IsFalse (board.ContainsPieceIn (position0));
			Assert.IsFalse (board.ContainsPieceIn (position1));
		}
		
		[Test]
		public void Execute_AddingPieces_AddsPieces ()
		{
			SetUpMoveActionToPutPieceIn (position0, piece0);
			SetUpMoveActionToPutPieceIn (position1, piece1);
			move = BuildMoveAction ();
			move.Execute (board);
			
			Assert.AreEqual (piece0, board.PieceIn (position0));
			Assert.AreEqual (piece1, board.PieceIn (position1));
		}
		
		[Test]
		[ExpectedException (typeof (InvalidOperationException),
		 ExpectedMessage = "Can't remove a piece from an empty location [3-2]")]
		public void Execute_RemovingPieceFromEmptyLocation_ThrowsException ()
		{
			position0 = new BoardPosition (3, 2);
			
			board.PutPieceIn (position0, piece0);
			SetUpMoveActionToRemovePieceIn (position0);
			board.RemovePieceIn (position0);
			
			move = BuildMoveAction ();
			
			move.Execute (board);
		}
		
		[Test]
		[ExpectedException (typeof (InvalidOperationException),
		 ExpectedMessage = "A different piece was expected in [3-2]")]
		public void Execute_RemovingWrongTypeOfPiece_ThrowsException ()
		{
			position0 = new BoardPosition (3, 2);
			
			board.PutPieceIn (position0, piece0);
			SetUpMoveActionToRemovePieceIn (position0);
			board.RemovePieceIn (position0);
			board.PutPieceIn (position0, piece1);
			
			move = BuildMoveAction ();
			
			move.Execute (board);
		}
		
		[Test]
		[ExpectedException (typeof (InvalidOperationException),
		 ExpectedMessage = "Can't add a piece to a non empty location [3-2]")]
		public void Execute_AddingPieceToNonEmptyLocation_ThrowsException ()
		{
			position0 = new BoardPosition (3, 2);
			board.PutPieceIn (position0, piece0);
			SetUpMoveActionToPutPieceIn (position0, piece0);
			
			move = BuildMoveAction ();
			
			move.Execute (board);
		}
		
		[Test]
		public void Undo_OnMoveRemovingPieces_RestoresPieces ()
		{
			board.PutPieceIn (position0, piece0);
			SetUpMoveActionToRemovePieceIn (position0);
			move = BuildMoveAction ();
			move.Execute (board);
			move.Undo (board);
			
			Assert.AreEqual (piece0, board.PieceIn (position0));
		}
		
		[Test]
		public void Undo_OnMoveAddingPieces_RemovesPieces ()
		{
			SetUpMoveActionToPutPieceIn (position0, piece0);
			move = BuildMoveAction ();
			move.Execute (board);
			move.Undo (board);
			
			Assert.IsFalse (board.ContainsPieceIn (position0));
		}
		
		[Test]
		[ExpectedException (typeof (InvalidOperationException),
		 ExpectedMessage = "Can't add a piece to a non empty location [3-2]")]
		public void Undo_MoveRemovingPieceOnCorruptedBoard_ThrowsException ()
		{
			board.PutPieceIn (position0, piece0);
			SetUpMoveActionToRemovePieceIn (position0);
			move = BuildMoveAction ();
			move.Execute (board);
			// Now board is empty in position 0
			
			board.PutPieceIn (position0, piece0);
			
			move.Undo (board); // Undoing moves should add piece to non empty location: exception thrown
		}
		
		[Test]
		public void Execute_ChangingPiece_WorksFine ()
		{
			board.PutPieceIn (position0, piece0);
			SetUpMoveActionToRemovePieceIn (position0);
			SetUpMoveActionToPutPieceIn (position0, piece2);
			move = BuildMoveAction ();
			move.Execute (board);
			
			Assert.IsTrue (board.ContainsPieceIn (position0));
			Assert.AreEqual (piece2, board.PieceIn (position0));
		}
		
		[Test]
		[ExpectedException]
		public void Execute_AddingTwoPiecesToSamePosition_ThrowsException ()
		{
			SetUpMoveActionToPutPieceIn (position0, piece0);
			SetUpMoveActionToPutPieceIn (position0, piece1);
			move = BuildMoveAction ();
			move.Execute (board);
		}
	}
}

