// 
// BoardController.cs
//  
// Author:
//       Matteo Nardi <matteo.nardi@gmx.com>
// 
// Copyright (c) 2011 Matteo Nardi
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Linq;

namespace Draughts.UI
{
	// Summary: This is the controller in the MVC pattern composed by BoardModel/BoardView/BoardController.
	//          It reacts to the view input by changing the model directly (setting square highlighting)
	//          or indirectly (changing the game class)
	public class BoardController
	{
		Game game;
		
		BoardView view;
		BoardModel model;
		
		bool isMoving;
		HumanPlayer movingPlayer;
		
		// List of positions clicked in the current move
		List <BoardPosition> clickedPositions = new List <BoardPosition> ();
		
		// List of allowed moves: each move is a list of positions from source to destination
		List <List <BoardPosition>> allowedMoves;
		
		public BoardController (Game game, BoardView widget)
		{
			this.game = game;
			this.view = widget;
			this.model = view.Model;
			
			TakeControlOfHumanPlayers ();
			
			// Take control of the widget
			view.BoardPositionClicked = OnBoardPositionClicked;
			
			// Retrive new moves only when move animations end.
			model.MoveAnimationEnded += delegate {
				if (game.Status == GameStatus.Stopped)
					GLibEventLoop.ExecuteOnIdle (game.RetriveNewMove); 
			};
		}
		
		void TakeControlOfHumanPlayers ()
		{
			if (game.WhitePlayer is HumanPlayer)
				((HumanPlayer) game.WhitePlayer).MoveRequested += OnMoveRequested;
			if (game.BlackPlayer is HumanPlayer)
				((HumanPlayer) game.BlackPlayer).MoveRequested += OnMoveRequested;
		}

		void OnBoardPositionClicked (BoardPosition position)
		{
			Console.WriteLine("Is moving " + isMoving.ToString());
			if (! isMoving) return;				
			
			// There are 4 possibilities:
			
			// 1) Is valid continuation of already started move
			if (clickedPositions.Count != 0 &&  GetPossibleContinuations ().Contains (position)) {
				clickedPositions.Add (position);
				
				if (IsValidMove ()) { 
					// Move found
					Move ();
				} else { 
					// Move not finished yet
					HighlightPath ();	
				}
			}
			
			// 2) Moving piece is clicked a second time ---> Move resetted
			else if (clickedPositions.Count > 0 && clickedPositions [0] == position) {
				model.ResetHightlight ();
				clickedPositions = new List <BoardPosition> ();
			}
			
			// 3) Moveable piece found in position ---> New move started
			else if (MovingPlayerContainsPieceIn (position) &&
			         allowedMoves.Any (move => move [0] == position)) {
				model.ResetHightlight ();
				clickedPositions = new List <BoardPosition> { position };
				HighlightPath ();
			}
			
			// 4) Useless click ---> Highlight all valid continuations with fade effect
			else {
				if (clickedPositions.Count == 0) 
					FadeHighlightContinuations (startingPiecesFadeColor);
				else
					FadeHighlightContinuations (destinationFadeColor);
			}
				
			model.Update ();
		}
		
		// This method returns the list of possible continuations.
		List <BoardPosition> GetPossibleContinuations ()
		{
			return allowedMoves
			    .Where (allowedMove => allowedMove.StartsWith (clickedPositions))
			    .Select (move => move [clickedPositions.Count])
				.Distinct ()
			    .ToList ();
		}
		
		bool MovingPlayerContainsPieceIn (BoardPosition position)
		{
			Color movingPlayerColor;
			if (movingPlayer == game.WhitePlayer)
				movingPlayerColor = Color.White;
			else if (movingPlayer == game.BlackPlayer)
				movingPlayerColor = Color.Black;
			else return false;
			
			return (game.Board.ContainsPieceIn (position)) &&
			       (game.Board.PieceIn (position).GetColor () == movingPlayerColor);
		}
		
		bool IsValidMove ()
		{
			return allowedMoves.Any (allowedMove => allowedMove.SequenceEqual (clickedPositions));
		}
		
		void Move ()
		{
			Console.WriteLine("moving");
			model.ResetHightlight ();
			isMoving = false;
			MoveDescription move = new MoveDescription (clickedPositions);
			clickedPositions = new List <BoardPosition> ();
			
			movingPlayer.Move (move);
		}
		
		void HighlightPath ()
		{
			if (clickedPositions.Count == 0)
				throw new InvalidOperationException ("Expected to operate on started move");
			
			model.ResetHightlight ();
			
			model.AddHighlight (clickedPositions [0], movingPieceColor);
			for (int i = 1; i < clickedPositions.Count; i++)
				model.AddHighlight (clickedPositions [i], visitedColor);
			
			foreach (BoardPosition continuation in GetPossibleContinuations ())
						model.AddHighlight (continuation, destinationColor);
		}
		
		void FadeHighlightContinuations (Cairo.Color color)
		{
			foreach (BoardPosition continuation in GetPossibleContinuations ())
					model.AddFadingHighlight (continuation, color, 500);
		}
		
		void OnMoveRequested (HumanPlayer player)
		{
			isMoving = true;
			movingPlayer = player;
			allowedMoves = game.MoveChecker.GetAllMoves ().ConvertAll (moveDescription => moveDescription.Steps);
			clickedPositions = new List <BoardPosition> ();
		}
		
		static readonly Cairo.Color movingPieceColor = new Cairo.Color (0,0.7,0);
		static readonly Cairo.Color startingPiecesFadeColor = new Cairo.Color (0.3,1,0.3, 1);
		static readonly Cairo.Color destinationColor = new Cairo.Color (0.6, 0, 0);
		static readonly Cairo.Color destinationFadeColor = new Cairo.Color (0.9, 0.3, 0.3, 0.4);
		static readonly Cairo.Color visitedColor = new Cairo.Color (0.4, 0, 0);
	}
}

static class ListExtensionMethods
{
	public static bool StartsWith <T> (this List <T> list, List <T> prefix) where T: IEquatable <T>
	{
		for (int i = 0; i < prefix.Count; i++)
			if (! list [i].Equals(prefix [i]))
				return false;
		return true;
	}
}
