// 
// BoardModel.cs
//  
// Author:
//       Matteo Nardi <matteo.nardi@gmx.com>
// 
// Copyright (c) 2011 Matteo Nardi
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Linq;
using System.Diagnostics;

namespace Draughts.UI
{
	// Summary: Contains the data (PieceModels and PositionModels) displayed by the BoardView widget.
	//          The list of pieces on the board is retrived from the game class.
	//          The list of highlighted positions is manipulated direcly by the controller.
	//          TODO: Supports for move animations.
	public class BoardModel
	{
		Queue <Action> pendingMoves;
		
		bool needsUpdate = false;
		bool isAnimating;
		public event Action MoveAnimationEnded;
		// This event is fired when the model has changed, it's used by the view to decide when to refresh
		public event Action Updated;
		
		public List <PieceModel> Pieces {
			get;
			private set;
		}
		
		// List of squares with a particular color.
		public List <PositionModel> HighlightedPositions {
			get;
			private set;
		}
		
		public GenericBoard <Draughts> Board {
			get {
				return game.Board;
			}
		}
		
		private Game game;
		
		public BoardModel (Game game)
		{
			this.game = game;
			this.isAnimating = false;
			game.NewMovePerformed += OnNewMove;
			
			LoadGamePieces ();
			ResetHightlight ();
			pendingMoves = new Queue <Action> ();
		}
		
		void LoadGamePieces ()
		{
			Pieces = new List <PieceModel> ();
			foreach (BoardPosition position in Board.AllPositions.Where (Board.ContainsPieceIn)) {
				var piece = new PieceModel (position, Board.PieceIn (position));
				Pieces.Add (piece);
				piece.AnimationEnded += OnPieceAnimationEnded;
			}
			
			needsUpdate = true;
		}
		
		public void ResetHightlight ()
		{	
			HighlightedPositions = new List <PositionModel> ();
			needsUpdate = true;
		}
		
		public void AddHighlight (BoardPosition position, Cairo.Color color)
		{
			var positionModel = new PositionModel (position, color);
			HighlightedPositions.Add (positionModel);
			
			needsUpdate = true;
		}
		
		// Set a special "fading" highlight in position for fadeTime milliseconds
		public void AddFadingHighlight (BoardPosition position, Cairo.Color color, int fadeTime)
		{
			var positionModel = new FadingPositionModel (position, color, fadeTime);
			HighlightedPositions.Add (positionModel);
			positionModel.AnimationEnded += CheckIfStillAnimating;
			
			StartAnimating ();
			needsUpdate = true;
		}
		
		void OnNewMove (MoveDescription move)
		{
			PieceModel movingPiece = Pieces.Find (piece => (piece.Position == move.Source));
			
			// Update piece type value since it might have been promoted
			Draughts newType = Board.PieceIn (move.Destination);
			
			List <PieceModel> removedPieces = 
				Pieces.Where (piece => (piece != movingPiece) && 
			                           (!Board.ContainsPieceIn (piece.Position))).ToList ();
			
			// We create an action to be executed later because we don't want overlapping animations.
			Action action = delegate {
				movingPiece.Move (move.Steps);
				movingPiece.Piece = newType;
				foreach (var piece in removedPieces)
					Pieces.Remove (piece);
				
				StartAnimating ();
				needsUpdate = true;
			};
						
			pendingMoves.Enqueue (action);
			if (Pieces.All (x => x.IsAnimating == false))
				DisplayNewMove ();
		}

		void StartAnimating ()
		{
			if (!isAnimating) {
				Console.WriteLine("Animating");
				isAnimating = true;
				GLib.Timeout.Add (10, Update);
			}
		}
		
		void OnPieceAnimationEnded ()
		{
			if (isAnimating) {
				if (Pieces.All (x => x.IsAnimating == false))
					if (MoveAnimationEnded != null)
						MoveAnimationEnded ();
				
				// Display new move when possible
				if (pendingMoves.Count > 0)
					DisplayNewMove ();
				
				CheckIfStillAnimating ();
			}
		}
		
		void CheckIfStillAnimating ()
		{
			if (HighlightedPositions.All (x => x.IsAnimating == false) &&
			    Pieces.All (x => x.IsAnimating == false))
			{
				StopAnimations ();
			}
		}
		
		// Fire the Updated event if any part of the model changed.
		// Return true if Updated was fired.
		public bool Update ()
		{
			Console.WriteLine("update");
			if (needsUpdate && Updated != null)
				Updated ();
			
			needsUpdate = isAnimating;
			return needsUpdate;
		}
	
		void DisplayNewMove ()
		{
			Action move = pendingMoves.Dequeue ();
			
			// We defer the move view action to idle time because it might change
			// the data our calling stack is depending on. 
			// (In particular it causes a crash when an eated piece is deleted)
			GLibEventLoop.ExecuteOnIdle (move);
		}
	
		void StopAnimations ()
		{
			isAnimating = false;
		}
	}
	
	public abstract class AnimatingModel {
		protected Stopwatch timer = new Stopwatch ();
		protected int animationDuration;
		
		public event Action AnimationEnded;
		public bool IsAnimating {
			get {
				if (timer.IsRunning && timer.ElapsedMilliseconds >= animationDuration) {
					timer.Stop ();
					if (AnimationEnded != null)
						AnimationEnded ();
				}
				
				return timer.IsRunning;
			}
		}
		
		protected void StartAnimating ()
		{
			timer.Reset ();
			timer.Start ();
		}
	}
	
	// Summary: Represents a piece on the board.
	public class PieceModel : AnimatingModel
	{
		List <BoardPosition> steps;
		
		public Draughts Piece;

		// This is the real position of the piece in the game.
		public BoardPosition Position {
			get;
			private set;
		}
		
		public PieceModel (BoardPosition position, Draughts piece)
		{
			this.Piece = piece;
			Position = position;
			animationDuration = 200;
		}
		
		public void Move (List <BoardPosition> steps)
		{
			Position = steps.Last ();
			this.steps = steps;
			
			StartAnimating ();
		}
		
		// Coordinates are expressed like this:
		// 0.0, it means the piece is centered on the first column, 
		// 1.0 -> centered on the second column, 
		// 0.5 -> the piece is on the line between the first and second column.
		public double CurrentX {
			get {
				return GetPosition (position => position.Column);
			}
		}
		
		public double CurrentY {
			get {
				return GetPosition (position => position.Row);
			}
		}
		
		double GetPosition (Func <BoardPosition, int> getCoordinate)
		{
			int elapsedTime = (int) timer.ElapsedMilliseconds;
			
			if (!IsAnimating) {
				return getCoordinate (Position);
			}
			
			double stepTime = animationDuration / (steps.Count -1);
			
			int completed = (int) (elapsedTime / stepTime);
			
			// This hack fixes some crash..maybe an of by one errors in the lines above? :/
			if (completed +1 >= steps.Count) completed = steps.Count -2;
			
			double fraction = (elapsedTime - stepTime * completed) / stepTime;
			int source = getCoordinate( steps [completed]);
			int dest = getCoordinate (steps [completed +1]);
			
			return source + (dest - source) * fraction;
		}
	}
	
	// Summary: Represents a square in the board with a special color.
	//          This color should be painted on top of the normal square color 
	//          (this allows trasparency effects)
	public class PositionModel : AnimatingModel
	{
		readonly public BoardPosition Position;
		
		protected Cairo.Color color;
		public virtual Cairo.Color Highlight {
			get {
				return color;
			}
		}
		
		public PositionModel (BoardPosition position, Cairo.Color highlight)
		{
			this.Position = position;
			this.color = highlight;
		}
	}
	
	// Summary: A special PositionModel with an animating highlight fading away with time
	public class FadingPositionModel : PositionModel
	{
		public override Cairo.Color Highlight {
			get {
				if (!IsAnimating) {
					return new Cairo.Color (0, 0, 0, 0);
				}

				return new Cairo.Color (color.R, color.G, color.B, GetAlpha ());
			}
		}
		
		double GetAlpha ()
		{
			double elapsedTime = timer.ElapsedMilliseconds;
			// We fade in and out: 
			// in 0 and totalTime this is 0, in totalTime/2 this is 1
			return color.A * (1.0 - Math.Pow ((elapsedTime - animationDuration/2) / (animationDuration/2), 2));
		}
		
		public FadingPositionModel (BoardPosition position, Cairo.Color highlight, int animationDuration)
			: base (position, highlight)
		{
			this.animationDuration = animationDuration;
			StartAnimating ();
		}
	}
}
