// 
// BoardView.cs
//  
// Author:
//       Matteo Nardi <matteo.nardi@gmx.com>
// 
// Copyright (c) 2011 Matteo Nardi
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Reflection;
using Gtk;
using Cairo;
using Draughts;

namespace Draughts.UI
{
	// Summary: This is the widget showing the board. 
	//          It's the view in a MVC pattern.
	//          It reads all the data from a BoardModel.
	//          It allows input (to the controller) in form of fired events.
	public class BoardView : DrawingArea
	{
		static readonly Cairo.Color DarkSquare = new Cairo.Color (0.6, 0.5, 0.5, 1);
		static readonly Cairo.Color LightSquare = new Cairo.Color (0.9, 0.9, 0.9, 1);
		static readonly Cairo.Color SurroundingSquare = new Cairo.Color (0.1, 0, 0, 1);
		static readonly Cairo.Color DarkPiece = new Cairo.Color (0, 0, 0, 1);
		static readonly Cairo.Color LightPiece = new Cairo.Color (1, 1, 1, 1);
		
		BoardModel model;
		
		// Window size
		int height, width;
		
		// Length of the edge of a single square in the board
		int edge; 
		
		// Position of the upper left corner
		int startingX, startingY;
		
		// Board size
		int rows, columns;		
		
		// The data being shown by the widget
		public BoardModel Model {
			get {
				return model;
			}
			set {
				model = value;
				rows = model.Board.NumberOfRows;
				columns = model.Board.NumberOfColumns;
				model.Updated += QueueDraw;
				UpdateBoardSize ();
				QueueDraw ();
			}
		}
		
		// This Action is fired when a square is clicked.
		// Note: this is a public Action and not an event to allow other classes
		//       to remove subscribers.  This makes it possible for a new BoardController to
		//       replace the old one, avoiding a (managed) memory leak.
		public Action <BoardPosition> BoardPositionClicked;
		
		public BoardView ()
		{	
			// This enables ButtonPressEvents, which are not enabled by default.
			AddEvents ((int) Gdk.EventMask.ButtonPressMask);
		}
		
		// Called when window is resized
		protected override bool OnConfigureEvent (Gdk.EventConfigure evnt)
		{
			evnt.Window.GetSize (out width, out height);
			if (model != null)
				UpdateBoardSize ();
			
			return true;
		}
		
		void UpdateBoardSize ()
		{
			edge = Math.Min (width / columns, height / rows) * 85/100;
			startingX = (width - (columns * edge)) / 2;
			startingY = (height - (rows * edge)) / 2;
		}
		
		// Called when widget should be refreshed
		protected override bool OnExposeEvent (Gdk.EventExpose evnt)
		{
			if (model != null) {
				using (Context cr = Gdk.CairoHelper.Create (evnt.Window)) {
					DrawSquareSurroundingBoard (cr);
					
					foreach (BoardPosition position in model.Board.AllPositions)
						DrawSquareInPosition (cr, position);
					
					foreach (PositionModel highlighted in model.HighlightedPositions)
						DrawSquareInPosition (cr, highlighted.Position, highlighted.Highlight);
					
					foreach (PieceModel piece in model.Pieces)
						DrawPiece (cr, piece); 
				}
			}
			return true;
		}
		
		protected override bool OnButtonPressEvent (Gdk.EventButton evnt)
		{
			if (model != null && evnt.Type == Gdk.EventType.ButtonPress) {
				if (evnt.Y < startingY || evnt.X < startingX) 
					return true;
				
				int selectedRow = rows - 1 - (int) ((evnt.Y - startingY) / edge);
				int selectedColumn = (int) ((evnt.X - startingX) / edge);
				BoardPosition selected = new BoardPosition (selectedRow, selectedColumn);
				
				if (model.Board.ContainsPosition (selected))
					BoardPositionClicked (selected);
			}
			return true;
		}

		void DrawSquareSurroundingBoard (Context cr)
		{
			cr.Save ();
			
			double offset = 0.5;
			cr.Rectangle (startingX - edge * offset, startingY - edge * offset,
			              edge * (columns + 2*offset),
			              edge * (rows + 2*offset));
			cr.Color = SurroundingSquare;
			cr.Fill ();
			
			cr.Restore ();
		}
		
		void DrawSquareInPosition (Context cr, BoardPosition position)
		{
			if (((Board) model.Board).IsDarkSquare (position)) 
				DrawSquareInPosition (cr, position, DarkSquare);
			else DrawSquareInPosition (cr, position, LightSquare);
		}
		
		void DrawSquareInPosition (Context cr, BoardPosition position, Cairo.Color color)
		{
			cr.Save ();
			
			cr.Rectangle (startingX + position.Column * edge, 
			              startingY + (rows - position.Row -1) * edge, 
			              edge, edge);
			
			cr.Color = color;
			cr.Fill ();
			
			cr.Restore ();
		}

		void DrawPiece (Context cr, PieceModel pieceModel)
		{
			cr.Save ();
		
			double centerX = startingX + (pieceModel.CurrentX + 0.5) * edge;
			double centerY = startingY + (rows - pieceModel.CurrentY - 0.5) * edge;
			double radius = edge * 0.4;
			double smallRadius = radius * 0.4;

			Draughts piece = pieceModel.Piece;
			
			// Draw outer circle
			cr.Arc (centerX, centerY, radius, 0, 2 * Math.PI);
			cr.StrokePreserve ();
							
			// Set color
			if (piece.GetColor () == Color.Black) 
				cr.Color = DarkPiece;
			else
				cr.Color = LightPiece;
			
			// Fill outer circle
			cr.Fill ();
			
			// Draw inner circle if it's a king
			if (piece.IsKing ()) {
				cr.MoveTo (centerX + smallRadius, centerY);
				cr.Arc (centerX, centerY, smallRadius, 0, 2 * Math.PI);
				cr.Color = new Cairo.Color (0.3, 0.3, 0.3, 1);
				cr.Fill ();
			}
		
			cr.Restore ();
		}
	}
}
