// 
// MainWindow.cs
//  
// Author:
//       Matteo Nardi <matteo.nardi@gmx.com>
// 
// Copyright (c) 2011 Matteo Nardi
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using Gtk;

namespace Draughts.UI
{
	public partial class MainWindow
	{
		// The game being displayed by the BoardView widget
		Game game;
		
		// The class used to build user interfaces from .glade files.
		// It's used to retrive the components.
		UserInterfaceCreator builder;
		
		BoardView boardWidget;
		
		Window window;
		NewGameDialog newGameDialog;
		
		public MainWindow ()
		{
			SetupWindow ();
			
			StartNewGame (PlayerType.Human, PlayerType.Medium);
			
			window.ShowAll ();
		}
		
		void SetupWindow ()
		{
			builder = new UserInterfaceCreator ("mainwindow.glade");
			
			window = (Window) builder.GetObject ("mainWindow");
			
			VBox layout = (VBox) builder.GetObject ("vbox1");
			boardWidget = new BoardView ();
			layout.Add (boardWidget);
			
			SetupActions ();
		}
		
		void SetupActions ()
		{
			MenuItem quitAction = (MenuItem) builder.GetObject ("quitMenuItem");
			MenuItem newGameAction = (MenuItem) builder.GetObject ("newGameMenuItem");
			MenuItem aboutAction = (MenuItem) builder.GetObject ("aboutMenuItem");
			
			window.DeleteEvent += delegate { Application.Quit (); };
			quitAction.Activated += delegate { Application.Quit (); };
			
			newGameAction.Activated += delegate { 
				if (newGameDialog == null)
					SetupNewGameDialog ();
				newGameDialog.Run (); 
			};
			
			aboutAction.Activated += delegate {
				var aboutDialog = new AboutDialog ();
				aboutDialog.Modal = true;
				aboutDialog.ShowAll ();
			};
		}
		
		void SetupNewGameDialog ()
		{
			newGameDialog = new NewGameDialog ();
			newGameDialog.NewGameRequested += StartNewGame;
		}
		
		void StartNewGame (PlayerType white, PlayerType black)
		{
			EndOldGame ();
			
			game = BuildGame (white, black);
			game.GameEnded += OnGameEnded;
			window.Title = "Checkers - " + white.ToString () + " vs " + black.ToString ();
			
			ConnectMVCComponents ();
			GLibEventLoop.ExecuteOnIdle (game.RetriveNewMove);
		}
		
		void EndOldGame ()
		{
			// Unsubscribe from old game events
			if (game != null)
				game.GameEnded -= OnGameEnded;
			
			game = null;
		}
		
		Game BuildGame (PlayerType white, PlayerType black)
		{
			Board board = Board.BuildDefault ();
			IPlayer whitePlayer = BuildPlayer (white, Color.White, board);
			IPlayer blackPlayer = BuildPlayer (black, Color.Black, board);
			
			Console.WriteLine (board.ToString ());
			
			return new Game (board, whitePlayer, blackPlayer);
		}
		
		// This function defines the AI for each difficulty level
		IPlayer BuildPlayer (PlayerType type, Color color, Board board)
		{
			if (type == PlayerType.VeryEasy)
				return new ComputerPlayer (color, board, 2, EvaluationFunctions.Random);
			
			else if (type == PlayerType.Easy)
				return new ComputerPlayer (color, board, 2, EvaluationFunctions.Basic);
				
			else if (type == PlayerType.Medium)
				return new ComputerPlayer (color, board, 4, EvaluationFunctions.Basic);
			
			else if (type == PlayerType.Hard)
				return new ComputerPlayer (color, board, 6, EvaluationFunctions.Promoting);
			
			// FIXME: this is not much better than Hard..
			else if (type == PlayerType.VeryHard)
				return new ComputerPlayer (color, board, 8, EvaluationFunctions.Promoting);

			else
				return new HumanPlayer ();
		}

		// Connects the model, view, controller components
		void ConnectMVCComponents ()
		{
			BoardModel model = new BoardModel (game);
			boardWidget.Model = model;
			new BoardController (game, boardWidget);
		}
		
		void OnGameEnded ()
		{
			string message;
			if (game.Status == GameStatus.WhiteWon)
				message = "White player has won!";
			else if (game.Status == GameStatus.BlackWon)
				message = "Black player has won!";
			else if (game.Status == GameStatus.Draw)
				message = "The game is drawn.";
			else throw new InvalidOperationException ("Game status corrupted");
			
			// Wait 100 milliseconds and show the end game dialog
			GLibEventLoop.ExecuteOnTimeout (100, ShowMessageDialog, message);	
		}
		
		void ShowMessageDialog (string message)
		{
			var dialog = new Gtk.MessageDialog (window, Gtk.DialogFlags.Modal, 
			                                    Gtk.MessageType.Info, Gtk.ButtonsType.Close,
			                                    message);
			dialog.ShowAll ();
			dialog.Run ();
			dialog.Destroy ();
		}
	}
	
	public enum PlayerType {
		Human,
		VeryEasy,
		Easy,
		Medium,
		Hard,
		VeryHard
	}
}

