// 
// Board.cs
//  
// Author:
//       Matteo Nardi <matteo.nardi@gmx.com>
// 
// Copyright (c) 2011 Matteo Nardi
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;

namespace Draughts
{
	// Summary: a container for Draughts pieces. 
	//          We customize GenericBoard behaviour by allowing pieces only on dark squares.
	public class Board : GenericBoard <Draughts>
	{
		public Board (int numberOfRows, int numberOfColumns) 
			: base (numberOfRows, numberOfColumns)
		{
		}
		
		public override void PutPieceIn (BoardPosition position, Draughts piece)
		{
			if (IsDarkSquare (position))
				base.PutPieceIn (position, piece);
			else
				throw new InvalidOperationException ("In Draughts Boards you are allowed to put pieces only on dark squares");
		}
		
		// This can be used to build a board with the default starting piece positioning.
		public static Board BuildDefault ()
		{
			int numberOfRows = 8;
			int numberOfColumns = 8;
			Draughts lowerPieceType = Draughts.WhiteMan;
			Draughts upperPieceType = Draughts.BlackMan;
			
			// All the rows but the two middle ones should be filled with pieces.
			int rows = numberOfRows / 2 -1;
			
			Board board = new Board (numberOfRows, numberOfColumns);
			foreach (BoardPosition position in board.AllPositions) {
				if (ItalianBoard.IsDarkSquare (position)) {
					if (position.Row < rows)
						board.PutPieceIn (position, lowerPieceType);
					if (position.Row >= rows+2)
						board.PutPieceIn (position, upperPieceType);
				}
			}
			return board;
		}
		
		public bool IsDarkSquare (BoardPosition position)
		{
			return ItalianBoard.IsDarkSquare (position);
		}
		
		static class ItalianBoard {
			public static bool IsDarkSquare (BoardPosition position)
			{
				// The bottom-left corner is white.
				return (position.Row + position.Column) % 2 == 1;
			}
		}
	}
	
	public static class BoardBuilder
	{
		public static Board BuildSquareBoard (string str)
		{
			int rows = (int) Math.Sqrt (str.Length);
			int columns = rows;
			
			Board board = new Board (rows, columns);
			
			foreach (BoardPosition position in board.AllPositions) {
				char symbol = str [(rows - 1 - position.Row) * columns + position.Column];
				switch (symbol) {
				case '.': // Empty position
					break;
				case 'w': // White man
					board.PutPieceIn (position, Draughts.WhiteMan);
					break;
				case 'W': // White king
					board.PutPieceIn (position, Draughts.WhiteKing);
					break;
				case 'b': // White man
					board.PutPieceIn (position, Draughts.BlackMan);
					break;
				case 'B': // White man
					board.PutPieceIn (position, Draughts.BlackKing);
					break;
				default: // Error
					throw new ArgumentException ("Unknown symbol");
				}
			}
			
			return board;
		}
	}
}

