// 
// BoardAnalysis.cs
//  
// Author:
//       Matteo Nardi <matteo.nardi@gmx.com>
// 
// Copyright (c) 2011 Matteo Nardi
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Linq;
using System.Collections.Generic;

namespace Draughts
{
	using MoveList = List <MoveDescription>;
	
	public interface IBoardAnalysis
	{
		bool IsValidMove (MoveDescription move);
		List <MoveDescription> GetAllMoves ();
		IMoveAction GetMoveAction (MoveDescription move);
	}
	
	// Summary: BoardAnalysis tells which moves are valid for a given Board configuration.
	//          Translates MoveDescriptions to MoveActions.
	// Design: the constructor needs a Board reference, it immediatly analyzes that board
	//         and stores all allowed MoveDescriptions (and respective MoveActions) in a Dictionary.
	//         This means that this class depends on the Board status only during construction.
	public class BoardAnalysis : IBoardAnalysis
	{
		GenericBoard <Draughts> board;
		Color movingPlayer;
		Dictionary <MoveDescription, MoveAction> allowedMoves;
		MoveLimitations limitations;
		
		public BoardAnalysis (GenericBoard <Draughts> board, Color movingPlayer)
		{
			this.board = board;
			this.movingPlayer = movingPlayer;
			this.allowedMoves = new Dictionary <MoveDescription, MoveAction> ();
			
			Analyze ();
		}
		
		public bool IsValidMove (MoveDescription move)
		{
			return allowedMoves.ContainsKey (move);
		}
		
		public List <MoveDescription> GetAllMoves ()
		{
			return allowedMoves.Keys.ToList ();
		}
		
		public IMoveAction GetMoveAction (MoveDescription move)
		{
			MoveAction action;
			if (allowedMoves.TryGetValue (move, out action))
				return action;
			else 
				throw new InvalidOperationException ("Can't get move action of unvalid move");
		}
		
		void Analyze ()
		{
			limitations = new MoveLimitations ();
			foreach (MoveDescription move in AllMoves ())
				if (SatisfiesConstraints (move))
					allowedMoves.Add (move, BuildMoveAction (move));
		}
		
		// This method returns all the moves which make sense, even if
		// they are not allowd (See SatisfiesConstraints for further details).
		// For each new move found, *limitations are updated.
		MoveList AllMoves ()
		{
			var moves = new MoveList ();
			
			foreach (BoardPosition source in board.AllPositions.Where (MovingPlayerHasPieceIn)) {
				var analyzer = new PieceAnalysis (source, board);
				analyzer.OnNewMoveFound = (move) => {
					moves.Add (move);
					limitations.Update (move, board);
				};
				analyzer.Run ();
			}
			return moves;
		}
		
		bool MovingPlayerHasPieceIn (BoardPosition position) {
			return board.ContainsPieceIn (position) &&
			       board.PieceIn (position).GetColor () == movingPlayer;
		}
		
		// Checks if the given move satisfies the game constraints.
		// For example (Draughts game) if the player can eat an opponent's piece,
		// he is not allowed to do a non-eating move.
		bool SatisfiesConstraints (MoveDescription move)
		{
			return limitations.IsValid (move, board);
		}
		
		MoveAction BuildMoveAction (MoveDescription move)
		{
			List <BoardPosition> toBeEmptied = new List <BoardPosition> ();
			List <Draughts> toBeRemoved = new List <Draughts> ();
		
			List <BoardPosition> toBeFilled  = new List <BoardPosition> ();
			List <Draughts> toBeAdded  = new List <Draughts> ();
			
			// Remove piece from source
			Draughts piece = board.PieceIn (move.Source);
			toBeEmptied.Add (move.Source);
			toBeRemoved.Add (piece);
			
			// Check for promotion
			if (move.Destination.Row == 0 && piece.GetColor () == Color.Black)
				piece = Draughts.BlackKing;
			if (move.Destination.Row == board.NumberOfRows -1 && piece.GetColor () == Color.White)
				piece = Draughts.WhiteKing;
			
			// Add piece to destination (and check for promotion!)
			toBeFilled.Add (move.Destination);
			toBeAdded.Add (piece);
			
			// Remove eated pieces
			if (move.IsEatingMove ())
				for (int i = 0; i < move.Steps.Count -1; i++) {
					BoardPosition source = move.Steps [i];
					BoardPosition destination = move.Steps [i+1];
					BoardPosition eated = new BoardPosition ((source.Row + destination.Row) / 2,
					                                         (source.Column + destination.Column) / 2);
					
					toBeEmptied.Add (eated);
					toBeRemoved.Add (board.PieceIn (eated));
				}
			
			return new MoveAction (toBeEmptied, toBeRemoved, toBeFilled, toBeAdded);
		}
	}
	
	// PieceAnalysis generates the list of move descriptions aviable starting
	// from a given position.
	class PieceAnalysis
	{
		// This method is executed everytime a valid move is found
		public Action <MoveDescription> OnNewMoveFound;
		
		readonly GenericBoard <Draughts> board;
		readonly BoardPosition source;
		
		readonly int forward;
		readonly Draughts movingPiece;
		readonly Color movingPlayer;
		
		public PieceAnalysis (BoardPosition source, GenericBoard <Draughts> board)
		{
			this.board = board;
			this.source = source;
			OnNewMoveFound = null;
			
			movingPiece = board.PieceIn (source);
			movingPlayer = movingPiece.GetColor ();
			
			// WARNING: We are assuming white pieces are in the lower part of the board, 
			// which is not true for all draghts types
			forward = movingPlayer == Color.White ? +1 : -1;
		}
		
		public void Run ()
		{
			if (OnNewMoveFound == null)
				throw new InvalidOperationException ("Piece Analysis was not told what to do when new moves are found");

			AddNonEatingMoves ();
			AddEatingMoves ();
		}

		void AddNonEatingMoves ()
		{
			List <BoardPosition> destinations = SpreadToDiagonalsBy (source, 1);
			
			foreach (BoardPosition destination in destinations) {
				if (board.ContainsPosition (destination) && board.IsEmptyIn(destination)) 
					OnNewMoveFound (new MoveDescription (source, destination));
			}
		}
		
		// Returns the list of locations the moving piece would reach moving diagonally by offset 
		List <BoardPosition> SpreadToDiagonalsBy (BoardPosition source, int offset)
		{
			var destinations = new List <BoardPosition> ();
			
			destinations.Add (new BoardPosition (source.Row + forward * offset, source.Column + offset));
			destinations.Add (new BoardPosition (source.Row + forward * offset, source.Column - offset));
			if (movingPiece.IsKing ()) {
				destinations.Add(new BoardPosition (source.Row - forward * offset, source.Column + offset));
				destinations.Add(new BoardPosition (source.Row - forward * offset, source.Column - offset));
			}
			// Return only the valid board positions
			return destinations.Where (board.ContainsPosition).ToList ();
		}

		void AddEatingMoves ()
		{
			var alreadyEaten = new HashSet <BoardPosition> ();
			var moveHistory = new List <BoardPosition> {source};
			Eat (source, alreadyEaten, moveHistory);
		}
		
		// Find all the eating moves the moving piece in *sourcePosition
		// would be able to do if all the pieces in *alreadyEaten were removed from the board.
		// *moveSteps contains the steps from *this.source to *sourcePosition
		void Eat (BoardPosition sourcePosition, 
		          HashSet <BoardPosition> alreadyEaten,
		          List <BoardPosition> moveSteps)
		{
			List <BoardPosition> destinations = SpreadToDiagonalsBy (sourcePosition, 2);
			
			foreach (BoardPosition destination in destinations) {
				
				int row = (destination.Row + sourcePosition.Row) /2;
				int column = (destination.Column + sourcePosition.Column) /2;
				BoardPosition eatedLocation = new BoardPosition (row, column);
				
				if ((board.IsEmptyIn (destination) || destination == source) &&
				    board.ContainsPieceIn (eatedLocation) &&
				    CanEat (movingPiece, board.PieceIn (eatedLocation)) &&
				    ! alreadyEaten.Contains (eatedLocation))
				{
					moveSteps.Add (destination);
					alreadyEaten.Add (eatedLocation);
					
					OnNewMoveFound (new MoveDescription (moveSteps));
					Eat (destination, alreadyEaten, moveSteps);
					
					alreadyEaten.Remove (eatedLocation);
					moveSteps.RemoveAt (moveSteps.Count -1);
				}
			}
		}
		
		bool CanEat (Draughts moving, Draughts eated)
		{
			// WARNING: We are assuming men can't eat kings. 
			return moving.GetColor () != eated.GetColor () &&
			       (moving.IsKing () || eated.IsMan ());
		}
	}
	
	// Italians move limitations
	class MoveLimitations
	{
		public int MinimumNumberOfCapturedPieces;
		public bool MustBeAKing;
		public int MinimumNumberOfCapturedKings;
		public int FirstKingOccurence;
		
		public MoveLimitations () 
		{
			SetNoLimitations ();
		}
		
		public MoveLimitations (MoveDescription move, GenericBoard <Draughts> board)
		{
			if (! move.IsEatingMove ()) 
				SetNoLimitations ();
			else {
				MinimumNumberOfCapturedPieces = move.Steps.Count - 1;
				MustBeAKing = board.PieceIn (move.Source).IsKing ();
				FindKings (move, board);
			}
		}
		
		void SetNoLimitations ()
		{
			MinimumNumberOfCapturedPieces = 0;
			MustBeAKing = false;
			MinimumNumberOfCapturedKings = 0;
			FirstKingOccurence = -1;
		}

		public bool IsValid (MoveDescription move, GenericBoard <Draughts> board)
		{
			MoveLimitations limitations = new MoveLimitations (move, board);
			
			return Compare (this, limitations) == 0;
		}
		
		public void Update (MoveDescription move, GenericBoard <Draughts> board) 
		{
			MoveLimitations newLimitations = new MoveLimitations (move, board);
			if (Compare (newLimitations, this) > 0) {
				MinimumNumberOfCapturedPieces = newLimitations.MinimumNumberOfCapturedPieces;
				MustBeAKing = newLimitations.MustBeAKing;
				MinimumNumberOfCapturedKings = newLimitations.MinimumNumberOfCapturedKings;
				FirstKingOccurence = newLimitations.FirstKingOccurence;
			}
		}
		
		void FindKings (MoveDescription move, GenericBoard <Draughts> board)
		{
			FirstKingOccurence = -1;
			MinimumNumberOfCapturedKings = 0;
			for (int i = 1; i < move.Steps.Count; i++) {
				BoardPosition eatedLocation = new BoardPosition (
				                  (move.Steps[i-1].Row + move.Steps[i].Row) / 2,
				                  (move.Steps[i-1].Column + move.Steps[i].Column) / 2);
				bool isKing = board.PieceIn (eatedLocation).IsKing ();
				if (isKing) {
					MinimumNumberOfCapturedKings++;
					if (FirstKingOccurence == -1)
						FirstKingOccurence = i;
				}
			}
		}
		
		static int Compare (MoveLimitations x, MoveLimitations y) {
			if (x.MinimumNumberOfCapturedPieces > y.MinimumNumberOfCapturedPieces) return 1;
			if (x.MinimumNumberOfCapturedPieces < y.MinimumNumberOfCapturedPieces) return -1;
			
			if (x.MustBeAKing && !y.MustBeAKing) return 1;
			if (!x.MustBeAKing && y.MustBeAKing) return -1;
			
			if (x.MinimumNumberOfCapturedKings > y.MinimumNumberOfCapturedKings) return 1;
			if (x.MinimumNumberOfCapturedKings < y.MinimumNumberOfCapturedKings) return -1;
			
			if (x.FirstKingOccurence < y.FirstKingOccurence) return 1;
			if (x.FirstKingOccurence > y.FirstKingOccurence) return -1;
			
			return 0;
		}
	}
}
