// 
// Game.cs
//  
// Author:
//       Matteo Nardi <matteo.nardi@gmx.com>
// 
// Copyright (c) 2011 Matteo Nardi
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;

namespace Draughts
{
	// TODO: implement swappable players.
	// TODO: keep track of move actions and support undo/redo.	
	
	public enum GameStatus {
		WaitingForMove,
		Stopped,
		WhiteWon,
		BlackWon,
		Draw
	}
	
	// Summary: Implements the game loop and acts as a container for the game components.
	public class Game
	{
		// Game starts in a Stopped status after construction.
		// When the class user calls RetriveNewMove, the status is switched to WaitingFromMove.
		// After a move is executed, the status is set to Stopped again.
		// When the game ends, Status is set to WhiteWin, BlackWin or Draw
		public GameStatus Status {
			get;
			private set;
		}
		
		// Analysis of the current board status. Tells which moves are valid.
		// Important: a new class is created every turn, don't keep references.
		public IBoardAnalysis MoveChecker {
			get;
			private set;
		}
		
		public Board Board {
			get;
			private set;
		}
		
		public IPlayer BlackPlayer {
			get;
			private set;
		}
		
		public IPlayer WhitePlayer {
			get;
			private set;
		}
				
		private Color movingPlayerColor;
		private IPlayer movingPlayer {
			get {
				if (movingPlayerColor == Color.Black)
					return BlackPlayer;
				else return WhitePlayer;
			}
		}
		
		public event Action <MoveDescription> NewMovePerformed;
		public event Action GameEnded;
		
		public Game (Board board, IPlayer white, IPlayer black)
		{
			this.Board = board;
			this.BlackPlayer = black;
			this.WhitePlayer = white;
			this.movingPlayerColor = Color.White;
			this.Status = GameStatus.Stopped;
			this.MoveChecker = GetBoardAnalysis ();
			
			BlackPlayer.MoveDecided += OnMoveDecided;
			WhitePlayer.MoveDecided += OnMoveDecided;
		}
		
		// Ask the moving player for a new move.
		public void RetriveNewMove ()
		{
			if (Status != GameStatus.Stopped)
				throw new InvalidOperationException ("Game can ask for a new move only when Status " +
				                                     "is set to Stopped, but was " + Status.ToString ());
			
			Status = GameStatus.WaitingForMove;
			movingPlayer.StartGetMove ();
		}
		
		void OnMoveDecided (IPlayer player, MoveDescription move)
		{
			if (Status != GameStatus.WaitingForMove)
				throw new InvalidOperationException ("We're not waiting for a move");
			
			if (player != movingPlayer)
				throw new InvalidOperationException ("Wrong player tried to move!");
			
			if (! MoveChecker.IsValidMove (move))
				throw new InvalidOperationException ("Player tried to perform illegal move!");
			
			Status = GameStatus.Stopped;
			ExecuteMove (move);
			
			movingPlayerColor = movingPlayerColor.SwitchColor ();
			MoveChecker = GetBoardAnalysis ();
			
			if (NewMovePerformed != null)
				NewMovePerformed (move);
			
			CheckIfGameEnded ();
		}
		
		void ExecuteMove (MoveDescription move)
		{
			UpdateTurnsSinceLastAdvancemet (move);
			
			IMoveAction action = MoveChecker.GetMoveAction (move);
			action.Execute (Board);
		}
		
		// Number of turns since last time a man (not king) advanced or 
		// a piece has been eaten. It's used to determinate game drawn.
		int turnsSinceLastAdvancement = 0;
		
		// Call this method before executing move, or it will crash
		void UpdateTurnsSinceLastAdvancemet (MoveDescription move)
		{
			if (move.IsEatingMove () || Board.PieceIn (move.Source).GetLevel () == Level.Man) 
				turnsSinceLastAdvancement = 0;
			
			turnsSinceLastAdvancement++;
		}
		
		void CheckIfGameEnded ()
		{
			// If no valid move is found, current player looses
			if (MoveChecker.GetAllMoves ().Count == 0) {
				if (movingPlayerColor == Color.White)
					Status = GameStatus.BlackWon;
				else Status = GameStatus.WhiteWon;
				
				GameEnded ();
			}
			
			// Draw
			if (turnsSinceLastAdvancement > 30) {
				Status = GameStatus.Draw;
				GameEnded ();
			}
		}
		
		protected virtual IBoardAnalysis GetBoardAnalysis ()
		{
			return new BoardAnalysis (Board, movingPlayerColor);
		}
	}
}

