// 
// Board.cs
//  
// Author:
//       Matteo Nardi <matteo.nardi@gmx.com>
// 
// Copyright (c) 2011 Matteo Nardi
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Linq;
using System.Collections.Generic;
namespace Draughts
{
	// Summary: the GenericBoard class is a generic container for pieces of different games.
	public class GenericBoard <Piece> : IEquatable <GenericBoard <Piece>> where Piece: struct
	{
		Piece? [,] board;
		
		public int NumberOfRows {
			get;
			private set;
		}
		
		public int NumberOfColumns {
			get;
			private set;
		}
		
		// This allows to iterate all valid BoardPositions using foreach
		public IEnumerable <BoardPosition> AllPositions {
			get {
				for (int row = 0; row < NumberOfRows; row++)
					for (int column = 0; column < NumberOfColumns; column++)
						yield return new BoardPosition (row, column);
			}
		}
		
		public GenericBoard (int numberOfRows, int numberOfColumns)
		{
			this.NumberOfRows = numberOfRows;
			this.NumberOfColumns = numberOfColumns;
			this.board = new Piece? [NumberOfRows, NumberOfColumns];
		}
		
		// TODO (MONO 2.10): use code contracts to assert input positions are valid 
		//                   (currently out of bound exceptions are thrown from the board array)
		public bool ContainsPieceIn (BoardPosition position)
		{
			return board [position.Row, position.Column].HasValue;
		}
		
		public bool IsEmptyIn (BoardPosition position)
		{
			return ! ContainsPieceIn (position);
		}
		
		public Piece PieceIn (BoardPosition position)
		{
			return board [position.Row, position.Column].Value;
		}
		
		public virtual void PutPieceIn (BoardPosition position, Piece piece)
		{
			if (! ContainsPieceIn (position))
				board [position.Row, position.Column] = piece;
			else 
				throw new InvalidOperationException ("Can't add piece to non empty position " + position.ToString ());
		}
		
		public void RemovePieceIn (BoardPosition position)
		{
			if (ContainsPieceIn (position))
				board [position.Row, position.Column] = null;
			else 
				throw new InvalidOperationException ("Can't remove piece from empty position " + position.ToString ());
		}
		
		public bool ContainsPosition (BoardPosition position)
		{
			return position.Column < this.NumberOfColumns &&
			       position.Column >= 0 &&
			       position.Row < this.NumberOfRows &&
			       position.Row >= 0;
		}
		
		public bool Equals (GenericBoard<Piece> other)
		{
			return this.NumberOfRows == other.NumberOfRows &&
			       this.NumberOfColumns == other.NumberOfColumns &&
			       AllPositions.All ((position) => 
					                  (this.board [position.Row, position.Column].Equals(other.board [position.Row, position.Column])));
		}
		
		public GenericBoard <Piece> Clone ()
		{
			GenericBoard <Piece> clone = new GenericBoard<Piece> (NumberOfRows, NumberOfColumns);
			
			foreach (BoardPosition position in AllPositions.Where (ContainsPieceIn))
					clone.PutPieceIn (position, PieceIn (position));
			
			return clone;
		}
	}
}
