// 
// ComputerPlayer.cs
//  
// Author:
//       Matteo Nardi <matteo.nardi@gmx.com>
// 
// Copyright (c) 2011 Matteo Nardi
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Linq;
namespace Draughts
{
	// Summary: An EvaluationFuncion analyzes the board passed as argument and returns 
	//          a numeric value denoting how good the situation is for player.
	//          These functions are important for AI algorithms such as MinMax.
	public delegate int EvaluationFuncion (GenericBoard <Draughts> board, Color player);
	
	// Summary: Implementation of the MinMax algorithm with alpha-beta pruning (negamax variant) 
	public class MinMaxAlgorithm
	{
		int analyzedStatusesCount;
		EvaluationFuncion evaluationFunction;
		int recursionLevel;
		
		
		public MinMaxAlgorithm (int depth, EvaluationFuncion evaluationFunction)
		{
			if (depth < 1)
				throw new InvalidOperationException ("MinMax recursion level must be > 0");
			
			this.evaluationFunction = evaluationFunction;
			this.recursionLevel = depth;
		}
		
		public MoveDescription GetBestMove (Board board, Color movingPlayer)
		{
			analyzedStatusesCount = 0;
			
			GenericBoard <Draughts> boardCopy = board.Clone ();
			BoardAnalysis boardAnalysis = new BoardAnalysis (boardCopy, movingPlayer);
			
			int infinity = int.MaxValue;
			int alpha = -infinity;
			int beta = infinity;
			MoveDescription bestMove = null;
			foreach (MoveDescription move in boardAnalysis.GetAllMoves ()) {
				IMoveAction action = boardAnalysis.GetMoveAction (move);
				action.Execute (boardCopy);
				
				int alpha_tmp = -GetBestMove (boardCopy, movingPlayer.SwitchColor (), 
				                              recursionLevel -1, -beta, -alpha);
				if (alpha_tmp > alpha || bestMove == null) {
					alpha = alpha_tmp;
					bestMove = move;
				}
				
				action.Undo (boardCopy);
			}
			
			Console.WriteLine ("Analyzed {0} board configurations.", analyzedStatusesCount);
			return bestMove;
		}
		
		int GetBestMove (GenericBoard <Draughts> board, Color movingPlayer, 
		                 int recursionLevel, int alpha, int beta)
		{
			analyzedStatusesCount++;
			
			if (recursionLevel == 0)
				return evaluationFunction (board, movingPlayer);
			
			BoardAnalysis analysis = new BoardAnalysis (board, movingPlayer);
			foreach (MoveDescription move in analysis.GetAllMoves ()) {
				IMoveAction action = analysis.GetMoveAction (move);
				action.Execute (board);
				alpha = Math.Max (alpha, -GetBestMove (board, movingPlayer.SwitchColor (), 
				                                       recursionLevel -1, -beta, -alpha));
				action.Undo (board);
				
				if (alpha >= beta) break;
			}
			return alpha;
		}
	}
	
	public static class EvaluationFunctions
	{
		static Random random = new Random (DateTime.Now.Millisecond);
		
		public static EvaluationFuncion Random = delegate { 
			return random.Next ();
		};
		
		// A decent valuation function for checkers, gives value to:
		// 1 - Number and kind of pieces left
		// 2 - A bonus for "aggressive" moves, makes the game a little more dynamic
		// 3 - Random component
		// If less then 5 pieces are left, special factors are considered.
		public static EvaluationFuncion Basic = delegate (GenericBoard <Draughts> board, Color player) {			
			int pieceValue = GetPieceValue (board, player);
			int numberOfPieces = board.AllPositions.Where (board.ContainsPieceIn).Count ();
			
			int endGameBonus = 0;
			if (numberOfPieces <= 5 && pieceValue > 0)
				endGameBonus = GetEndGameBonus (board, player);
			
			return pieceValue * 10000
			       + pieceValue * 10000 / numberOfPieces
			       + endGameBonus * 100
			       + GetAggressiveBonus (board, player) * 10
			       + random.Next () % 10;
		};
		
		// This is an improved Basic evaluation function which gives value to
		// pieces which are close to become king
		public static EvaluationFuncion Promoting = delegate (GenericBoard <Draughts> board, Color player) {
			return Basic (board, player)
			       + GetAboutToBeCrownedValue (board, player) * 100;
		};
		
		static int GetPieceValue (GenericBoard <Draughts> board, Color player)
		{
			int pieceValue = 0;
			
			foreach (BoardPosition position in board.AllPositions.Where (board.ContainsPieceIn)) {
				Draughts piece = board.PieceIn (position);
				
				int sign = (player == piece.GetColor ()) ? +1 : -1;
				if (piece.GetLevel () == Level.King)
					pieceValue += 5 * sign;
				else
					pieceValue += 3 * sign;
			}
			
			return pieceValue;
		}
		
		// Return the value associated with the pieces close to become king.
		static int GetAboutToBeCrownedValue (GenericBoard <Draughts> board, Color player)
		{
			int crowningValue = 0;
			
			foreach (BoardPosition position in board.AllPositions.Where (board.ContainsPieceIn)) {
				Draughts piece = board.PieceIn (position);
				if (piece.GetLevel () == Level.King)
					continue;
				
				// FIXME: we must get an easy way to determinate in which direction
				// the current player moves.
				bool forward = piece.GetColor () == Color.White;
				
				// We get a bonus for being close to the the opponent side 
				// if we are in the first or last column (since we can't be eaten)
				// or if we have no opponent in front of us in our columns
				bool crowning = 
				   position.Column == 0 ||
				   position.Column == board.NumberOfColumns -1 ||
				   board.AllPositions
				   .Where (p => (forward ? p.Row > position.Row : p.Row < position.Row) && // in front of us..
						        Math.Abs (position.Column - p.Column) <= 1) // and in our columns..
				   .All (p => board.ContainsPieceIn (p) == false ||
				              board.PieceIn (p).GetColor () == piece.GetColor ());
				
				if (crowning) {
					// How distant we are from our side: longer is better because
					// we are close be become kings
					int distance = forward ?
					               position.Row :
					               board.NumberOfRows -1 -position.Row;
					int sign = (player == piece.GetColor ()) ? +1 : -1;
					crowningValue += distance * sign;
				}
			}
			
			return crowningValue;
		}
		
		// Give a bonus to kings close to opponent pieces
		static int GetAggressiveBonus (GenericBoard <Draughts> board, Color player)
		{
			var playerPositions = board.AllPositions
			                           .Where (board.ContainsPieceIn)
			                           .Where (p => board.PieceIn (p).GetColor () == player);
			var opponentPositions = board.AllPositions
			                             .Where (board.ContainsPieceIn)
			                             .Where (p => board.PieceIn (p).GetColor () != player);
			
			return -playerPositions
			           .Where (p => board.PieceIn (p).GetLevel () == Level.King)
			           .Sum (p => opponentPositions.Sum (pos => Distance (p, pos)));
		}
		
		// Special function for game endings, it makes sense for the player winning the game.
		// It tries to keep opponent pieces away from the safe corners and tries to block
		// opponent pieces (pieces which can't move without being eaten)
		static int GetEndGameBonus (GenericBoard <Draughts> board, Color player)
		{
			var corners = new List <BoardPosition> {
				new BoardPosition (0, 1),
				new BoardPosition (1, 0),
				new BoardPosition (board.NumberOfRows -2, board.NumberOfColumns -1),
				new BoardPosition (board.NumberOfRows -1, board.NumberOfColumns -2),
			};
			
			var playerPositions = board.AllPositions
			                           .Where (board.ContainsPieceIn)
			                           .Where (p => board.PieceIn (p).GetColor () == player);
			var opponentPositions = board.AllPositions
			                             .Where (board.ContainsPieceIn)
			                             .Where (p => board.PieceIn (p).GetColor () != player);
			
			Func <BoardPosition, bool> isBlockedPosition = delegate (BoardPosition p) {
				return playerPositions.Any (pos => (DistanceFromBorder (pos, board) == 2 &&
			                                        Distance (pos, p) == 2));
			};
			
			int blockedPieces = opponentPositions
			                       .Where (p => DistanceFromBorder (p, board) == 0)
			                       .Where (isBlockedPosition).Count ();
			int safePieces = opponentPositions.Where (p => corners.Contains (p)).Count ();
			
			return blockedPieces - safePieces;
		}
		
		static int Distance (BoardPosition p1, BoardPosition p2)
		{
			return Math.Abs (p1.Row - p2.Row) + 
			       Math.Abs (p1.Column - p2.Column);
		}
		
		static int DistanceFromBorder (BoardPosition position, GenericBoard <Draughts> board)
		{
			return Math.Min (Math.Min (board.NumberOfRows -1 -position.Row, position.Row),
			                 Math.Min (board.NumberOfColumns -1 -position.Column, position.Column));
			
		}		
	}
}
