// 
// MoveAction.cs
//  
// Author:
//       Matteo Nardi <matteo.nardi@gmx.com>
// 
// Copyright (c) 2011 Matteo Nardi
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE. 

using System;
using System.Collections.Generic;

namespace Draughts
{	
	public interface IMoveAction
	{
		void Execute (GenericBoard <Draughts> board);
		void Undo (GenericBoard <Draughts> board);
	}
	
	// Summary: MoveAction represents an action executed on a Board.
	//          It supports undo/redo of removing and adding Draughts pieces.
	//          The constructor expects to recive four lists indicating what to add and what to remove.
	//          Pieces are first removed and then added: to change a piece in a given position, remove
	//          the old one and put there the new one. 
	//          If board is not valid an exception is thrown, but the board might become corrupted.
	// Design: through dependency injection, MoveAction is not coupled with any specific board:
	//         nothing prevents calling Execute on different boards, more then one time on the same one,
	//         or calling Undo on a board on which Execute was never called.
	public class MoveAction : IMoveAction
	{
		// Board positions emptied when action is executed
		readonly List <BoardPosition> locationsToBeEmptied;
		
		// Pieces expected to be removed when action is executed
		// removedPieces[i] is removed from locationsToBeEmptied[i] on execution, it's added on undo
		readonly List <Draughts> removedPieces;
		
		// Board positions where a piece is added when action is executed
		readonly List <BoardPosition> locationsToBeFilled;
		
		// Pieces added when action is executed
		// addedPieces[i] is added to locationsToBeFilled [i] on execution, it's removed on undo
		readonly List <Draughts> addedPieces;
		
		public MoveAction (List <BoardPosition> locationsToBeEmptied,
		                   List <Draughts> removedPieces,
		                   List <BoardPosition> locationsToBeFilled,
		                   List <Draughts> addedPieces)
		{
			if (locationsToBeEmptied.Count != removedPieces.Count)
				throw new ArgumentException ("locationsToBeEmptied has different number of elements then removedPieces");
			if (locationsToBeFilled.Count != addedPieces.Count)
				throw new ArgumentException ("locationsToBeFilled has different number of elements then addedPieces");
			
			this.locationsToBeEmptied = locationsToBeEmptied;
			this.removedPieces = removedPieces;
			this.locationsToBeFilled = locationsToBeFilled;
			this.addedPieces = addedPieces;
		}
		
		public void Execute (GenericBoard <Draughts> board)
		{
			AssertLocationsAreNonEmpty (locationsToBeEmptied, removedPieces, board);			
			RemovePiecesFrom (locationsToBeEmptied, board);
			
			AssertLocationsAreEmpty (locationsToBeFilled, board);
			AddPiecesTo (locationsToBeFilled, addedPieces, board);
		}
		
		public void Undo (GenericBoard <Draughts> board)
		{
			AssertLocationsAreNonEmpty (locationsToBeFilled, addedPieces, board);			
			RemovePiecesFrom (locationsToBeFilled, board);
			
			AssertLocationsAreEmpty (locationsToBeEmptied, board);
			AddPiecesTo (locationsToBeEmptied, removedPieces, board);
		}
		
		void AssertLocationsAreNonEmpty (List <BoardPosition> positions, 
		                                 List <Draughts> expectedPieces,
		                                 GenericBoard <Draughts> board)
		{
			for (int i = 0; i < positions.Count; i++) {
				BoardPosition location = positions [i];
				Draughts piece = expectedPieces [i];
				
				if (! board.ContainsPieceIn (location))
					throw new InvalidOperationException ("Can't remove a piece from an empty location " + location.ToString ());
				if (board.PieceIn (location) != piece)
					throw new InvalidOperationException ("A different piece was expected in " + location.ToString ());
			}
		}
		
		void AssertLocationsAreEmpty (List <BoardPosition> positions,
		                              GenericBoard <Draughts> board)
		{
			foreach (BoardPosition position in positions)
				if (board.ContainsPieceIn (position))
					throw new InvalidOperationException ("Can't add a piece to a non empty location " + position.ToString ());
		}
		
		void RemovePiecesFrom (List <BoardPosition> positions, GenericBoard <Draughts> board)
		{
			foreach (BoardPosition position in positions)
				board.RemovePieceIn (position);	
		}
		
		void AddPiecesTo (List <BoardPosition> destinations, 
		                  List<Draughts> pieces, GenericBoard <Draughts> board)
		{
			for (int i = 0; i < destinations.Count; i++)
				board.PutPieceIn (destinations [i], pieces [i]);
		}
	}
}

